from fastmolwidget.viewer_widget import MoleculeViewerWidget
from fastmolwidget.molecule2D import MoleculeWidget
from fastmolwidget.loader import MoleculeLoader

__all__ = ["MoleculeViewerWidget", "MoleculeWidget", "MoleculeLoader"]


def main() -> None:
    print("Hello from fastmolwidget!")
